/* valasymbol.vala
 *
 * Copyright (C) 2006-2007  Jürg Billeter
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
 * Author:
 * 	Jürg Billeter <j@bitron.ch>
 */

#include <vala/valasymbol.h>
#include <gee/readonlycollection.h>
#include <vala/valascope.h>

struct _ValaSymbolPrivate {
	char* _name;
	gboolean _active;
	ValaSymbolAccessibility _access;
	gboolean _is_imported;
	ValaScope* _owner;
	ValaScope* _scope;
};
#define VALA_SYMBOL_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), VALA_TYPE_SYMBOL, ValaSymbolPrivate))
enum  {
	VALA_SYMBOL_DUMMY_PROPERTY,
	VALA_SYMBOL_PARENT_SYMBOL,
	VALA_SYMBOL_OWNER,
	VALA_SYMBOL_NAME,
	VALA_SYMBOL_ACTIVE,
	VALA_SYMBOL_ACCESS,
	VALA_SYMBOL_SCOPE,
	VALA_SYMBOL_IS_IMPORTED
};
static char* vala_symbol_real_get_cprefix (ValaSymbol* self);
static char* vala_symbol_real_get_lower_case_cname (ValaSymbol* self, const char* infix);
static char* vala_symbol_real_get_lower_case_cprefix (ValaSymbol* self);
static GeeCollection* vala_symbol_real_get_cheader_filenames (ValaSymbol* self);
static GObject * vala_symbol_constructor (GType type, guint n_construct_properties, GObjectConstructParam * construct_properties);
static gpointer vala_symbol_parent_class = NULL;
static void vala_symbol_dispose (GObject * obj);


/**
 * Check if this symbol is just internal API (and therefore doesn't need 
 * to be listed in header files for instance) by traversing parent symbols
 * and checking their accessibility.
 */
gboolean vala_symbol_is_internal_symbol (ValaSymbol* self) {
	g_return_val_if_fail (VALA_IS_SYMBOL (self), FALSE);
	{
		ValaSymbol* sym;
		ValaSymbol* _tmp3;
		ValaSymbol* _tmp2;
		ValaSymbol* _tmp1;
		ValaSymbol* _tmp0;
		sym = NULL;
		_tmp3 = NULL;
		_tmp2 = NULL;
		_tmp1 = NULL;
		_tmp0 = NULL;
		for (sym = (_tmp1 = (_tmp0 = self, (_tmp0 == NULL ? NULL : g_object_ref (_tmp0))), (sym == NULL ? NULL : (sym = (g_object_unref (sym), NULL))), _tmp1); NULL != sym; sym = (_tmp3 = (_tmp2 = vala_symbol_get_parent_symbol (sym), (_tmp2 == NULL ? NULL : g_object_ref (_tmp2))), (sym == NULL ? NULL : (sym = (g_object_unref (sym), NULL))), _tmp3)) {
			if (VALA_SYMBOL_ACCESSIBILITY_PRIVATE == vala_symbol_get_access (sym)) {
				gboolean _tmp4;
				return (_tmp4 = TRUE, (sym == NULL ? NULL : (sym = (g_object_unref (sym), NULL))), _tmp4);
			}
		}
		(sym == NULL ? NULL : (sym = (g_object_unref (sym), NULL)));
	}
	return FALSE;
}


/**
 * Returns the fully expanded name of this symbol for use in
 * human-readable messages.
 *
 * @return full name
 */
char* vala_symbol_get_full_name (ValaSymbol* self) {
	char* _tmp3;
	gboolean _tmp4;
	char* _tmp7;
	char* _tmp8;
	g_return_val_if_fail (VALA_IS_SYMBOL (self), NULL);
	if (vala_symbol_get_parent_symbol (self) == NULL) {
		const char* _tmp0;
		_tmp0 = NULL;
		return (_tmp0 = vala_symbol_get_name (self), (_tmp0 == NULL ? NULL : g_strdup (_tmp0)));
	}
	if (vala_symbol_get_name (self) == NULL) {
		return vala_symbol_get_full_name (vala_symbol_get_parent_symbol (self));
	}
	_tmp3 = NULL;
	if ((_tmp4 = (_tmp3 = vala_symbol_get_full_name (vala_symbol_get_parent_symbol (self))) == NULL, (_tmp3 = (g_free (_tmp3), NULL)), _tmp4)) {
		const char* _tmp5;
		_tmp5 = NULL;
		return (_tmp5 = vala_symbol_get_name (self), (_tmp5 == NULL ? NULL : g_strdup (_tmp5)));
	}
	_tmp7 = NULL;
	_tmp8 = NULL;
	return (_tmp8 = g_strdup_printf ("%s.%s", (_tmp7 = vala_symbol_get_full_name (vala_symbol_get_parent_symbol (self))), vala_symbol_get_name (self)), (_tmp7 = (g_free (_tmp7), NULL)), _tmp8);
}


/**
 * Returns the camel case string to be prepended to the name of members
 * of this symbol when used in C code.
 *
 * @return the camel case prefix to be used in C code
 */
static char* vala_symbol_real_get_cprefix (ValaSymbol* self) {
	g_return_val_if_fail (VALA_IS_SYMBOL (self), NULL);
	if (vala_symbol_get_name (self) == NULL) {
		return g_strdup ("");
	} else {
		const char* _tmp1;
		_tmp1 = NULL;
		return (_tmp1 = vala_symbol_get_name (self), (_tmp1 == NULL ? NULL : g_strdup (_tmp1)));
	}
}


char* vala_symbol_get_cprefix (ValaSymbol* self) {
	return VALA_SYMBOL_GET_CLASS (self)->get_cprefix (self);
}


/**
 * Returns the C name of this symbol in lower case. Words are
 * separated by underscores. The lower case C name of the parent symbol
 * is prefix of the result, if there is one.
 *
 * @param infix a string to be placed between namespace and data type
 *              name or null
 * @return      the lower case name to be used in C code
 */
static char* vala_symbol_real_get_lower_case_cname (ValaSymbol* self, const char* infix) {
	g_return_val_if_fail (VALA_IS_SYMBOL (self), NULL);
	return NULL;
}


char* vala_symbol_get_lower_case_cname (ValaSymbol* self, const char* infix) {
	return VALA_SYMBOL_GET_CLASS (self)->get_lower_case_cname (self, infix);
}


/**
 * Returns the string to be prefixed to members of this symbol in
 * lower case when used in C code.
 *
 * @return      the lower case prefix to be used in C code
 */
static char* vala_symbol_real_get_lower_case_cprefix (ValaSymbol* self) {
	g_return_val_if_fail (VALA_IS_SYMBOL (self), NULL);
	return g_strdup ("");
}


char* vala_symbol_get_lower_case_cprefix (ValaSymbol* self) {
	return VALA_SYMBOL_GET_CLASS (self)->get_lower_case_cprefix (self);
}


/**
 * Returns a list of C header filenames users of this symbol must
 * include.
 *
 * @return list of C header filenames for this symbol
 */
static GeeCollection* vala_symbol_real_get_cheader_filenames (ValaSymbol* self) {
	g_return_val_if_fail (VALA_IS_SYMBOL (self), NULL);
	return GEE_COLLECTION (gee_read_only_collection_new (g_strdup, g_free, NULL));
}


GeeCollection* vala_symbol_get_cheader_filenames (ValaSymbol* self) {
	return VALA_SYMBOL_GET_CLASS (self)->get_cheader_filenames (self);
}


/**
 * Converts a string from CamelCase to lower_case.
 *
 * @param camel_case a string in camel case
 * @return           the specified string converted to lower case
 */
char* vala_symbol_camel_case_to_lower_case (const char* camel_case) {
	GString* result;
	const char* i;
	gboolean first;
	const char* _tmp0;
	char* _tmp1;
	g_return_val_if_fail (camel_case != NULL, NULL);
	result = g_string_new ("");
	i = camel_case;
	first = TRUE;
	while (g_utf8_strlen (i, -1) > 0) {
		gunichar c;
		c = g_utf8_get_char (i);
		if (g_unichar_isupper (c) && !first) {
			const char* t;
			gboolean prev_upper;
			gboolean next_upper;
			/* current character is upper case and
			 * we're not at the beginning */
			t = g_utf8_prev_char (i);
			prev_upper = g_unichar_isupper (g_utf8_get_char (t));
			t = g_utf8_next_char (i);
			next_upper = g_unichar_isupper (g_utf8_get_char (t));
			if (!prev_upper || (g_utf8_strlen (i, -1) >= 2 && !next_upper)) {
				gint len;
				/* previous character wasn't upper case or
				 * next character isn't upper case*/
				len = g_utf8_strlen (result->str, -1);
				if (len != 1 && g_utf8_get_char (g_utf8_offset_to_pointer (result->str, len - 2)) != '_') {
					/* we're not creating 1 character words */
					g_string_append_c (result, '_');
				}
			}
		}
		g_string_append_unichar (result, g_unichar_tolower (c));
		first = FALSE;
		i = g_utf8_next_char (i);
	}
	_tmp0 = NULL;
	_tmp1 = NULL;
	return (_tmp1 = (_tmp0 = result->str, (_tmp0 == NULL ? NULL : g_strdup (_tmp0))), (result == NULL ? NULL : (result = (g_string_free (result, TRUE), NULL))), _tmp1);
	(result == NULL ? NULL : (result = (g_string_free (result, TRUE), NULL)));
}


ValaSymbol* vala_symbol_get_parent_symbol (ValaSymbol* self) {
	g_return_val_if_fail (VALA_IS_SYMBOL (self), NULL);
	if (vala_symbol_get_owner (self) == NULL) {
		return NULL;
	} else {
		return vala_scope_get_owner (vala_symbol_get_owner (self));
	}
}


ValaScope* vala_symbol_get_owner (ValaSymbol* self) {
	g_return_val_if_fail (VALA_IS_SYMBOL (self), NULL);
	return self->priv->_owner;
}


void vala_symbol_set_owner (ValaSymbol* self, ValaScope* value) {
	g_return_if_fail (VALA_IS_SYMBOL (self));
	self->priv->_owner = value;
	vala_scope_set_parent_scope (self->priv->_scope, value);
}


char* vala_symbol_get_name (ValaSymbol* self) {
	g_return_val_if_fail (VALA_IS_SYMBOL (self), NULL);
	return self->priv->_name;
}


void vala_symbol_set_name (ValaSymbol* self, const char* value) {
	char* _tmp2;
	const char* _tmp1;
	g_return_if_fail (VALA_IS_SYMBOL (self));
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_name = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL ? NULL : g_strdup (_tmp1))), (self->priv->_name = (g_free (self->priv->_name), NULL)), _tmp2);
}


gboolean vala_symbol_get_active (ValaSymbol* self) {
	g_return_val_if_fail (VALA_IS_SYMBOL (self), FALSE);
	return self->priv->_active;
}


void vala_symbol_set_active (ValaSymbol* self, gboolean value) {
	g_return_if_fail (VALA_IS_SYMBOL (self));
	self->priv->_active = value;
}


ValaSymbolAccessibility vala_symbol_get_access (ValaSymbol* self) {
	g_return_val_if_fail (VALA_IS_SYMBOL (self), 0);
	return self->priv->_access;
}


void vala_symbol_set_access (ValaSymbol* self, ValaSymbolAccessibility value) {
	g_return_if_fail (VALA_IS_SYMBOL (self));
	self->priv->_access = value;
}


ValaScope* vala_symbol_get_scope (ValaSymbol* self) {
	g_return_val_if_fail (VALA_IS_SYMBOL (self), NULL);
	return self->priv->_scope;
}


gboolean vala_symbol_get_is_imported (ValaSymbol* self) {
	g_return_val_if_fail (VALA_IS_SYMBOL (self), FALSE);
	return self->priv->_is_imported;
}


void vala_symbol_set_is_imported (ValaSymbol* self, gboolean value) {
	g_return_if_fail (VALA_IS_SYMBOL (self));
	self->priv->_is_imported = value;
}


static GObject * vala_symbol_constructor (GType type, guint n_construct_properties, GObjectConstructParam * construct_properties) {
	GObject * obj;
	ValaSymbolClass * klass;
	GObjectClass * parent_class;
	ValaSymbol * self;
	klass = VALA_SYMBOL_CLASS (g_type_class_peek (VALA_TYPE_SYMBOL));
	parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
	obj = parent_class->constructor (type, n_construct_properties, construct_properties);
	self = VALA_SYMBOL (obj);
	{
		ValaScope* _tmp0;
		_tmp0 = NULL;
		self->priv->_scope = (_tmp0 = vala_scope_new (self), (self->priv->_scope == NULL ? NULL : (self->priv->_scope = (g_object_unref (self->priv->_scope), NULL))), _tmp0);
		vala_symbol_set_active (self, TRUE);
	}
	return obj;
}


static void vala_symbol_get_property (GObject * object, guint property_id, GValue * value, GParamSpec * pspec) {
	ValaSymbol * self;
	self = VALA_SYMBOL (object);
	switch (property_id) {
		case VALA_SYMBOL_PARENT_SYMBOL:
		g_value_set_object (value, vala_symbol_get_parent_symbol (self));
		break;
		case VALA_SYMBOL_OWNER:
		g_value_set_object (value, vala_symbol_get_owner (self));
		break;
		case VALA_SYMBOL_NAME:
		g_value_set_string (value, vala_symbol_get_name (self));
		break;
		case VALA_SYMBOL_ACTIVE:
		g_value_set_boolean (value, vala_symbol_get_active (self));
		break;
		case VALA_SYMBOL_ACCESS:
		g_value_set_int (value, vala_symbol_get_access (self));
		break;
		case VALA_SYMBOL_SCOPE:
		g_value_set_object (value, vala_symbol_get_scope (self));
		break;
		case VALA_SYMBOL_IS_IMPORTED:
		g_value_set_boolean (value, vala_symbol_get_is_imported (self));
		break;
		default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}


static void vala_symbol_set_property (GObject * object, guint property_id, const GValue * value, GParamSpec * pspec) {
	ValaSymbol * self;
	self = VALA_SYMBOL (object);
	switch (property_id) {
		case VALA_SYMBOL_OWNER:
		vala_symbol_set_owner (self, g_value_get_object (value));
		break;
		case VALA_SYMBOL_NAME:
		vala_symbol_set_name (self, g_value_get_string (value));
		break;
		case VALA_SYMBOL_ACTIVE:
		vala_symbol_set_active (self, g_value_get_boolean (value));
		break;
		case VALA_SYMBOL_ACCESS:
		vala_symbol_set_access (self, g_value_get_int (value));
		break;
		case VALA_SYMBOL_IS_IMPORTED:
		vala_symbol_set_is_imported (self, g_value_get_boolean (value));
		break;
		default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}


static void vala_symbol_class_init (ValaSymbolClass * klass) {
	vala_symbol_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (ValaSymbolPrivate));
	G_OBJECT_CLASS (klass)->get_property = vala_symbol_get_property;
	G_OBJECT_CLASS (klass)->set_property = vala_symbol_set_property;
	G_OBJECT_CLASS (klass)->constructor = vala_symbol_constructor;
	G_OBJECT_CLASS (klass)->dispose = vala_symbol_dispose;
	VALA_SYMBOL_CLASS (klass)->get_cprefix = vala_symbol_real_get_cprefix;
	VALA_SYMBOL_CLASS (klass)->get_lower_case_cname = vala_symbol_real_get_lower_case_cname;
	VALA_SYMBOL_CLASS (klass)->get_lower_case_cprefix = vala_symbol_real_get_lower_case_cprefix;
	VALA_SYMBOL_CLASS (klass)->get_cheader_filenames = vala_symbol_real_get_cheader_filenames;
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_SYMBOL_PARENT_SYMBOL, g_param_spec_object ("parent-symbol", "parent-symbol", "parent-symbol", VALA_TYPE_SYMBOL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_SYMBOL_OWNER, g_param_spec_object ("owner", "owner", "owner", VALA_TYPE_SCOPE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_SYMBOL_NAME, g_param_spec_string ("name", "name", "name", NULL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_SYMBOL_ACTIVE, g_param_spec_boolean ("active", "active", "active", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_SYMBOL_ACCESS, g_param_spec_int ("access", "access", "access", G_MININT, G_MAXINT, 0, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_SYMBOL_SCOPE, g_param_spec_object ("scope", "scope", "scope", VALA_TYPE_SCOPE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_SYMBOL_IS_IMPORTED, g_param_spec_boolean ("is-imported", "is-imported", "is-imported", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
}


static void vala_symbol_init (ValaSymbol * self) {
	self->priv = VALA_SYMBOL_GET_PRIVATE (self);
}


static void vala_symbol_dispose (GObject * obj) {
	ValaSymbol * self;
	self = VALA_SYMBOL (obj);
	(self->priv->_name = (g_free (self->priv->_name), NULL));
	(self->priv->_scope == NULL ? NULL : (self->priv->_scope = (g_object_unref (self->priv->_scope), NULL)));
	G_OBJECT_CLASS (vala_symbol_parent_class)->dispose (obj);
}


GType vala_symbol_get_type (void) {
	static GType vala_symbol_type_id = 0;
	if (G_UNLIKELY (vala_symbol_type_id == 0)) {
		static const GTypeInfo g_define_type_info = { sizeof (ValaSymbolClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) vala_symbol_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (ValaSymbol), 0, (GInstanceInitFunc) vala_symbol_init };
		vala_symbol_type_id = g_type_register_static (VALA_TYPE_CODE_NODE, "ValaSymbol", &g_define_type_info, G_TYPE_FLAG_ABSTRACT);
	}
	return vala_symbol_type_id;
}




